from typing import Dict, Mapping, Optional, Union


class GqlError(Exception):
    def __init__(
        self,
        body: Optional[Union[Dict, str]] = None,
        headers: Optional[Mapping] = None,
        message: str = "",
        status_code: int = 0,
        summary: str = "",
    ):
        self.body = body
        self.headers = headers
        self.message = message
        self.status_code = status_code
        self.summary = summary
        self.retryable = status_code >= 500
        super(GqlError, self).__init__()

    def __str__(self) -> str:
        return self.summary


class InvalidSessionError(Exception):
    pass


class InvalidConfigFileError(Exception):
    pass



from configparser import NoSectionError
from typing import Any
from unittest import TestCase
from unittest.mock import Mock, call, patch

from pycarlo.common.errors import InvalidSessionError
from pycarlo.core import Session


class SessionTest(TestCase):
    @patch.object(Session, "_read_config")
    def test_session_with_params(self, mock_read: Mock):
        mcd_id, mcd_token = "foo", "bar"
        session = Session(mcd_id=mcd_id, mcd_token=mcd_token)

        self.assertEqual(session.id, mcd_id)
        self.assertEqual(session.token, mcd_token)
        mock_read.assert_not_called()

    @patch("pycarlo.core.session.MCD_DEFAULT_API_ID", "foo")
    @patch("pycarlo.core.session.MCD_DEFAULT_API_TOKEN", "bar")
    @patch.object(Session, "_read_config")
    def test_session_with_env(self, mock_read: Mock):
        session = Session()

        self.assertEqual(session.id, "foo")
        self.assertEqual(session.token, "bar")
        mock_read.assert_not_called()

    @patch("pycarlo.core.session.MCD_DEFAULT_API_TOKEN", "bar")
    @patch.object(Session, "_read_config")
    def test_session_with_mixed(self, mock_read: Mock):
        mcd_id = "foo"
        session = Session(mcd_id=mcd_id)

        self.assertEqual(session.id, mcd_id)
        self.assertEqual(session.token, "bar")
        mock_read.assert_not_called()

    @patch("pycarlo.core.session.MCD_DEFAULT_API_TOKEN", "qux")
    @patch.object(Session, "_read_config")
    def test_session_with_precedence(self, mock_read: Mock):
        mcd_id, mcd_token = "foo", "bar"
        session = Session(mcd_id=mcd_id, mcd_token=mcd_token)

        self.assertEqual(session.id, mcd_id)
        self.assertEqual(session.token, mcd_token)
        mock_read.assert_not_called()

    @patch.object(Session, "_read_config")
    def test_session_with_partial(self, mock_read: Mock):
        with self.assertRaises(InvalidSessionError):
            Session(mcd_id="foo")
        mock_read.assert_not_called()

    @patch("pycarlo.core.session.configparser")
    def test_read_config(self, mock_parser: Mock):
        mcd_id, mcd_token, mcd_api_endpoint, mcd_config_path = "foo", "bar", "endpoint", "path/"
        mock_parser.ConfigParser().get.side_effect = [mcd_id, mcd_token, mcd_api_endpoint]

        session = Session(mcd_config_path=mcd_config_path)
        mock_parser.assert_has_calls = [
            call.ConfigParser(),
            call.ConfigParser().read("path/profiles.ini"),
            call.ConfigParser().get("default", "mcd_id"),
            call.ConfigParser().get("default", "mcd_token"),
            call.ConfigParser().get("default", "mcd_api_endpoint"),
        ]  # type: ignore
        self.assertEqual(session.id, mcd_id)
        self.assertEqual(session.token, mcd_token)
        self.assertEqual(session.endpoint, mcd_api_endpoint)

    @patch.object(Session, "_get_config_parser")
    def test_read_config_with_bad_section(self, mock_parser: Mock):
        class InvalidParser:
            def read(self, *args: Any, **kwargs: Any):
                pass

            def get(self, *args: Any, **kwargs: Any):
                raise NoSectionError("")

        mock_parser.return_value = InvalidParser()

        with self.assertRaises(InvalidSessionError):
            Session()

    @patch("pycarlo.core.session.pkg_resources")
    @patch("pycarlo.core.session.uuid")
    def test_set_session_name(self, mock_uuid: Mock, mock_pkg: Mock):
        mcd_id, mcd_token = "foo", "bar"
        mock_uuid_val, mock_pkg_val = "42", "99"

        mock_uuid.uuid4.return_value = mock_uuid_val
        mock_pkg.get_distribution().version = mock_pkg_val
        session = Session(mcd_id=mcd_id, mcd_token=mcd_token)

        self.assertEqual(session.id, mcd_id)
        self.assertEqual(session.token, mcd_token)
        self.assertEqual(session.session_name, f"python-sdk-{mock_pkg_val}-{mock_uuid_val}")

    def test_set_session_endpoint(self):
        mcd_id, mcd_token, endpoint = "foo", "bar", "test.com"
        self.assertEqual(
            Session(mcd_id=mcd_id, mcd_token=mcd_token, endpoint=endpoint).endpoint, endpoint
        )
        self.assertEqual(
            Session(mcd_id=mcd_id, mcd_token=mcd_token).endpoint,
            "https://api.getmontecarlo.com/graphql",
        )



from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Dict, Optional, Union, cast
from uuid import UUID

from dataclasses_json import LetterCase, dataclass_json

from pycarlo.common import get_logger, http
from pycarlo.common.files import BytesFileReader, JsonFileReader, to_path
from pycarlo.core import Client, Query
from pycarlo.features.dbt.queries import (
    GET_DBT_UPLOAD_URL,
    SEND_DBT_ARTIFACTS_EVENT,
)
from pycarlo.features.pii import PiiFilterer, PiiService
from pycarlo.features.user import UserService

logger = get_logger(__name__)


class InvalidArtifactsException(Exception):
    pass


class InvalidFileFormatException(Exception):
    pass


@dataclass_json(letter_case=LetterCase.CAMEL)  # type: ignore
@dataclass
class DbtArtifacts:
    manifest: str
    run_results: str
    logs: Optional[str]


class DbtImporter:
    """
    Import dbt run artifacts to Monte Carlo
    """

    DEFAULT_PROJECT_NAME = "default-project"
    DEFAULT_JOB_NAME = "default-job"

    def __init__(
        self,
        mc_client: Optional[Client] = None,
        user_service: Optional[UserService] = None,
        pii_service: Optional[PiiService] = None,
        print_func: Callable = logger.info,
    ):
        self._mc_client = mc_client or Client()
        self._user_service = user_service or UserService(mc_client=self._mc_client)
        self._pii_service = pii_service or PiiService(mc_client=self._mc_client)
        self._print_func = print_func
        self._pii_filterer = self._init_pii_filterer()

    def import_run(
        self,
        manifest_path: Union[Path, str],
        run_results_path: Union[Path, str],
        logs_path: Optional[Union[Path, str]] = None,
        project_name: str = DEFAULT_PROJECT_NAME,
        job_name: str = DEFAULT_JOB_NAME,
        resource_id: Optional[Union[str, UUID]] = None,
    ):
        """
        Import artifacts from a single dbt command execution.

        :param manifest_path: local path to the dbt manifest file (manifest.json)
        :param run_results_path: local path to the dbt run results file (run_results.json)
        :param logs_path: local path to a file containing dbt run logs
        :param project_name: Project name (perhaps a logical group of dbt models, analogous to a
                             project in dbt Cloud)
        :param job_name: Job name (perhaps a logical sequence of dbt commands, analogous to a
                         job in dbt Cloud)
        :param resource_id: identifier of a Monte Carlo resource (warehouse or lake) to use to
                            resolve dbt models to tables, this will be required if you have more
                            than one
        """
        # get resource
        resource = self._user_service.get_resource(resource_id)

        # read local artifacts
        manifest = JsonFileReader(manifest_path).read()
        run_results = JsonFileReader(run_results_path).read()
        logs = BytesFileReader(logs_path).read() if logs_path else None

        # extract dbt invocation id (and verify it is the same for each artifact)
        invocation_id = self._get_invocation_id(
            manifest_path=manifest_path,
            manifest=manifest,
            run_results_path=run_results_path,
            run_results=run_results,
        )

        # upload artifacts to S3 (using pre-signed URLs)
        artifacts = DbtArtifacts(
            manifest=self._upload_artifact(
                project_name=project_name,
                invocation_id=invocation_id,
                file_path=to_path(manifest_path),
                content=manifest,
            ),
            run_results=self._upload_artifact(
                project_name=project_name,
                invocation_id=invocation_id,
                file_path=to_path(run_results_path),
                content=run_results,
            ),
            logs=self._upload_artifact(
                project_name=project_name,
                invocation_id=invocation_id,
                file_path=to_path(logs_path),  # type: ignore
                content=logs,
            )
            if logs
            else None,
        )

        # publish event indicating run artifacts are ready for processing
        self._mc_client(
            query=SEND_DBT_ARTIFACTS_EVENT,
            variables=dict(
                projectName=project_name,
                jobName=job_name,
                invocationId=invocation_id,
                artifacts=artifacts.to_dict(),  # type: ignore
                resourceId=str(resource.id),
            ),
        )

        self._print_func("Finished sending run artifacts to Monte Carlo")

    def _get_invocation_id(
        self,
        manifest_path: Union[Path, str],
        manifest: Dict,
        run_results_path: Union[Path, str],
        run_results: Dict,
    ) -> str:
        manifest_invocation_id = self._extract_invocation_id(path=manifest_path, data=manifest)
        run_results_invocation_id = self._extract_invocation_id(
            path=run_results_path, data=run_results
        )

        if manifest_invocation_id != run_results_invocation_id:
            raise InvalidArtifactsException(
                "dbt invocation ids do not match between manifest and run results files"
            )

        return manifest_invocation_id

    @staticmethod
    def _extract_invocation_id(path: Union[Path, str], data: Dict) -> str:
        try:
            return data["metadata"]["invocation_id"]
        except KeyError:
            raise InvalidArtifactsException(
                f"Unable to get dbt invocation id from '{path}'. Unexpected file format"
            )

    def _upload_artifact(
        self,
        project_name: str,
        invocation_id: str,
        file_path: Path,
        content: Union[bytes, str, Dict],
    ) -> str:
        self._print_func(f"Uploading {file_path.name}...")
        http.upload(
            method="put",
            url=self._get_presigned_url(
                project_name=project_name, invocation_id=invocation_id, file_name=file_path.name
            ),
            content=self._pii_filterer.filter_content(content),
        )
        return file_path.name

    def _get_presigned_url(self, project_name: str, invocation_id: str, file_name: str) -> str:
        response = cast(
            Query,
            self._mc_client(
                query=GET_DBT_UPLOAD_URL,
                variables=dict(
                    projectName=project_name, invocationId=invocation_id, fileName=file_name
                ),
            ),
        )

        return cast(str, response.get_dbt_upload_url)

    def _init_pii_filterer(self):
        pii_filters = self._pii_service.get_pii_filters_config()
        return PiiFilterer(filters_config=pii_filters)



from unittest import TestCase

import requests
import responses
import responses.registries
from requests import Timeout

from pycarlo.common.retries import Backoff, retry_with_backoff


class MockBackoff(Backoff):
    def __init__(self, count: int):
        super(MockBackoff, self).__init__(start=0, maximum=0)
        self.count = count

    def backoff(self, attempt: int) -> float:
        return 0

    def delays(self):
        for i in range(self.count):
            yield 0


class RetryDecoratorTest(TestCase):
    @responses.activate(registry=responses.registries.OrderedRegistry)
    def test_exhaustion_after_retry(self):
        url = "https://example.com"
        responses.get(url=url, body=Timeout())
        responses.get(url=url, body=Timeout())
        responses.get(url=url, body=Timeout())

        backoff = MockBackoff(2)

        @retry_with_backoff(backoff=backoff, exceptions=Timeout)
        def call_api():
            r = requests.get(url=url)
            return r.text

        with self.assertRaises(expected_exception=Timeout) as e:
            call_api()
        self.assertIsInstance(obj=e.exception, cls=Timeout)
        responses.mock.assert_call_count(url=url, count=3)

    @responses.activate(registry=responses.registries.OrderedRegistry)
    def test_success_after_retry(self):
        url = "https://example.com"
        responses.get(url=url, body=Timeout())
        responses.get(url=url, body=Timeout())
        responses.get(url=url, body="hello")

        backoff = MockBackoff(2)

        @retry_with_backoff(backoff=backoff, exceptions=Timeout)
        def call_api():
            r = requests.get(url=url)
            return r.text

        self.assertEqual("hello", call_api())
        responses.mock.assert_call_count(url=url, count=3)

    @responses.activate(registry=responses.registries.OrderedRegistry)
    def test_success_without_retry(self):
        url = "https://example.com"
        responses.get(url=url, body="hello")

        backoff = MockBackoff(2)

        @retry_with_backoff(backoff=backoff, exceptions=Timeout)
        def call_api():
            r = requests.get(url=url)
            return r.text

        self.assertEqual("hello", call_api())
        responses.mock.assert_call_count(url=url, count=1)



from dataclasses import dataclass, field
from typing import Any, Callable, Dict, Optional

from dataclasses_json import dataclass_json

from pycarlo.features.metadata import AllowBlockList, FilterEffectType, FilterType, MetadataFilter


@dataclass_json
@dataclass
class MetadataFiltersContainer:
    """
    More documentation and samples in the link below:
    https://www.notion.so/montecarlodata/Catalog-Schema-Filtering-59edd6eff7f74c94ab6bfca75d2e3ff1

    MetadataFiltersContainer class that includes a metadata_filters list that works
    in the following way:
    A list of filters where:
    - each filter can be a block or allow filter
    - each filter can optionally filter: project, dataset, table and table_type
    - each filter matches using of the following types: exact match, regular expression, prefix
    - there's a default effect (allow/block) configured in the list, those elements with no matching
      filter will be resolved with the default effect in the list.
    - This class supports filtering objects in memory or generating SQL conditions for filtering,
      for SQL generation an encoder function is required that maps the different filter types to
      SQL, also a dictionary mapping from property name to column name is required (for example to
      map 'project' to 'database' or 'dataset' to 'schema'.
    - The order in which elements are added to the list is not relevant, priority is assigned based
      on the effect, rules with the default effect have the higher priority and stop the iteration.

    Filtering works by prioritizing "explicit" filters, a filter is considered explicit when it is
    configured with the same effect used as the default in the list.
    Let's suppose we have the following sample data:
    - prj_1: ds_1, ds_2
    - prj_2: ds_1, ds_2, ds_3
    - project_3: dataset_1, ds_2
    _ project_4: dataset_4

    These are examples using the sample data above:
    - list(default=allow): block(prj_*), allow(prj_1)
        - will allow prj_1 and all those projects not matching prj_*
        - allowed: (prj_1, all datasets), (project_3, all datasets), (project_4, all datasets)
    - list(default=block): allow(prj_*), block(prj_1)
        - will allow only prj_* except prj_1 that is explicitly blocked
        - allowed: (prj_2, all datasets)
    - list(default=allow): allow(*), block(prj_1)
        - will allow everything, allow(*) is considered a explicit rule and has the highest priority
        - allowed: everything
    - list(default=allow): block(*), allow(prj_1)
        - will allow only prj_1
        - allowed: (prj_1, all datasets)
    - list(default=allow): block(*, ds_*)
        - will block all datasets named ds_*
        - allowed: (project_3, dataset_1), (project_4, all datasets)
    - list(default=allow): allow(prj_2, ds_3), block(*, ds_*)
        - will block all datasets named ds_3 except for ds_3 in prj_2 that is explicitly allowed.
          Please note order is not relevant and the result would be exactly the same for:
            block(*, ds_*), allow(prj_2, ds_3)
        - allowed: (prj_2, ds_3), (project_3, dataset_1), (project_4, all datasets)
    - list(default=block): allow(*, dataset_*):
        - only datasets named dataset_* will be allowed
        - allowed: (project_3, dataset_1), (project_4, dataset_4)
    """

    metadata_filters: AllowBlockList = field(default_factory=AllowBlockList)

    @property
    def is_metadata_filtered(self) -> bool:
        return bool(self.metadata_filters.filters)

    @property
    def is_metadata_blocked(self):
        """
        Helper method for detecting an edge case where everything is blocked because the default
        effect is block and all filters are also blocking, in this case it doesn't make sense to
        run queries or filter data.
        """
        return self.metadata_filters.default_effect == FilterEffectType.BLOCK and all(
            f.effect == FilterEffectType.BLOCK for f in self.metadata_filters.filters
        )

    def is_project_with_datasets_filtered(self, project: str) -> bool:
        """
        Returns True if there's at least one filter configured for the specified project filtering
        on datasets, this can be used to check if get_sql_conditions for the project is going to
        return an empty query or not.
        """
        return self.is_metadata_filtered and any(
            f.matches(project=project) and f.dataset is not None
            for f in self.metadata_filters.filters
        )

    def is_whole_project_blocked(self, project: str) -> bool:
        """
        Helper method to be used when projects are iterated first, returns True if the project is
        fully blocked (blocked by a filter with no dataset or blocked by default) False otherwise.
        For example for a list like: list(default=allow): block(prj_1, ds_1) this method will return
        False as prj_1 is not fully blocked, there might be a (prj_1, ds_2) that is allowed, so
        prj_1 still needs to be iterated.
        """
        # the condition parameter below is to include blocking filters only if applied to the
        # whole project, we don't want to exclude a project just because a dataset is excluded
        effect = self._get_effect(
            metadata_filters=self.metadata_filters,
            force_regexp=False,
            condition=lambda f: not f.dataset or f.effect == FilterEffectType.ALLOW,
            project=project,
        )
        return effect == FilterEffectType.BLOCK

    def is_metadata_element_allowed(self, **kwargs: Any) -> bool:
        """
        Metadata elements filtering, iterates all filters looking for a match, if there's
        a match in a filter with the default effect, it's considered an explicit filter
        and search stops with the result effect being the default one.
        If there's a match in a filter not configured with the default effect, search continues.
        When the search completes, if there was a match then the result effect will be the one in
        the matched filter (must be the non-default effect).
        If there was no match the result effect will be the default one.
        Result for this method is True only if the result effect is ALLOW.

        Data is matched using properties specified in kwargs, the following keys are supported
        in kwargs: 'project', 'dataset', 'table', 'table_type'.
        """
        effect = self._get_effect(
            metadata_filters=self.metadata_filters, force_regexp=False, **kwargs
        )
        return effect == FilterEffectType.ALLOW

    @staticmethod
    def _get_effect(
        metadata_filters: AllowBlockList,
        force_regexp: bool,
        condition: Optional[Callable[[MetadataFilter], bool]] = None,
        **kwargs: Any,
    ) -> FilterEffectType:
        """
        Returns the effect for a metadata element with the properties specified by kwargs
        (project, dataset, table, table_type).
        If there's an explicit filter matching (a filter is explicit if the effect is the default
        one) then default effect is returned.
        If there's a match in the "other effect" list then the "other effect" is returned, we're
        calling "other effect" to the effect that is not the default one.
        If no matching filter, the default effect is returned.
        """
        if not metadata_filters.filters or any(
            f.matches(force_regexp, **kwargs)
            for f in metadata_filters.get_default_effect_filters(condition=condition)
        ):
            return metadata_filters.default_effect

        if any(
            f.matches(force_regexp, **kwargs)
            for f in metadata_filters.get_other_effect_filters(condition=condition)
        ):
            return metadata_filters.other_effect

        return metadata_filters.default_effect

    def is_dataset_allowed(self, project: Optional[str], dataset: str) -> bool:
        """
        Helper method intended to be used when projects and datasets are iterated in memory.
        It returns True if the dataset in the given project is allowed (not blocked), this is
        equivalent to call is_metadata_element_allowed(project=project, dataset=dataset)
        """
        return self.is_metadata_element_allowed(project=project, dataset=dataset)

    def get_sql_conditions(
        self,
        column_mapping: Dict,  # maps project and dataset to the column name to use
        encoder: Callable[[str, str, FilterType], str],
        project: Optional[str] = None,
    ) -> Optional[str]:
        """
        Helper method that returns a SQL query fragment with conditions for the current filters.
        If project is specified this will return conditions only for the specified project and this
        is supposed to be called after checking that is_project_with_datasets_filtered and
        is_project_allowed returned True.
        column_mapping is used to map filter fields (like project and dataset) to the actual
        database columns (like database and schema).
        encoder is used to encode a filter in the SQL dialect, it needs to encode to expressions
        like "database = 'db_1'" or "database REGEXP 'db_.*'".
        Examples:
            - default=block, filters=allow(project=x_*), block(project=x_1), allow(project=z)
                SQL: NOT(project='x_1') AND ((project REGEXP 'x_*') OR (project='z')
            - default=allow, filters=block(project=x_*), allow(project=x_1), block(project=z)
                SQL: project='x_1' OR (NOT(project REGEXP 'x_*') AND NOT(project='z'))
        Basically we first put all filters matching the default condition joined by
        AND/OR (block/allow), and then all filters with the other effect joined by OR/AND.
        """
        if not self.metadata_filters.filters:
            return None
        if project and not self.is_project_with_datasets_filtered(project):
            return None

        def project_condition(f: MetadataFilter):
            return not project or f.matches(project=project)

        default_effect = self.metadata_filters.default_effect
        default_effect_filters = self.metadata_filters.get_default_effect_filters(
            condition=project_condition
        )
        other_effect_filters = self.metadata_filters.get_other_effect_filters(
            condition=project_condition
        )
        default_effect_op = " OR " if default_effect == FilterEffectType.ALLOW else " AND "
        other_effect_op = " AND " if default_effect == FilterEffectType.ALLOW else " OR "

        default_effect_conditions = default_effect_op.join(
            [
                f"({self._get_sql_field_condition(f, column_mapping, encoder)})"
                for f in default_effect_filters
            ]
        )
        other_effect_conditions = other_effect_op.join(
            [
                self._get_sql_field_condition(f, column_mapping, encoder)
                for f in other_effect_filters
            ]
        )
        conditions = default_effect_conditions
        if conditions and other_effect_conditions:
            conditions += default_effect_op
            conditions += "(" + other_effect_conditions + ")"
        elif not conditions:
            conditions = other_effect_conditions
        return f"({conditions})" if conditions else ""

    @staticmethod
    def _get_sql_field_condition(
        mf: MetadataFilter,
        column_mapping: Dict,
        encoder: Callable[[str, str, FilterType], str],
    ):
        # The comparison is performed case-insensitive (check MetadataFilter._safe_match)
        # We can use LOWER here since it is part of standard SQL (like AND/OR/NOT), so including it
        # here is a way to make sure that all comparisons are case-insensitive in the SQL sentences
        # for all engines
        conditions = " AND ".join(
            [
                encoder(
                    f"LOWER({column})",
                    getattr(mf, field).lower(),
                    mf.type if field == mf.filter_type_target_field() else FilterType.EXACT_MATCH,
                )
                for (field, column) in column_mapping.items()
                if getattr(mf, field) is not None
            ]
        )
        return f"NOT({conditions})" if mf.effect == FilterEffectType.BLOCK else f"({conditions})"



import random
import time
from abc import ABC, abstractmethod
from functools import wraps
from typing import Any, Callable, Generator, Tuple, Type, Union


class Backoff(ABC):
    """
    Backoff is an abstract class that dictates a retry strategy.
    It contains an abstract method `backoff` that returns a calculated delay based on the provided
    attempt.
    """

    def __init__(self, start: float, maximum: float):
        """
        Defines a new backoff retry strategy.

        :param start: the scaling factor for any calculated delays.
        :param maximum: defines a cap on the calculated delays to prevent prohibitively long waits
                        that could time out.
        """
        self.start = start
        self.maximum = maximum

    @abstractmethod
    def backoff(self, attempt: int) -> float:
        pass

    def delays(self) -> Generator[float, None, None]:
        """
        Generates a duration of time to delay for each successive call based on the configured
        backoff strategy.

        :return: a generator that yields the next delay duration.
        """
        duration = 0
        retries = 0
        while duration < self.maximum:
            duration = self.backoff(retries)
            yield duration
            retries += 1


def retry_with_backoff(
    backoff: Backoff, exceptions: Union[Type[Exception], Tuple[Type[Exception], ...]]
) -> Callable:
    """
    A decorator to retry a function based on the :param:`backoff` provided
    if any of the provided :param:`exceptions` are raised.

    :param backoff: the retry strategy to employ.
    :param exceptions: the exceptions that should trigger the retry. Can be further customized by
                       defining a custom attribute `retryable` on the exception class. The retries
                       are abandoned If retryable returns False.
    :return: The same result the decorated function returns.
    """

    def _retry(func: Callable):
        @wraps(func)
        def _impl(*args: Any, **kwargs: Any):
            delays = backoff.delays()
            while True:
                try:
                    result = func(*args, **kwargs)
                except exceptions as exception:
                    if not hasattr(exception, "retryable") or exception.retryable:  # type: ignore
                        pass
                    else:
                        raise exception

                    try:
                        delay = next(delays)
                    except StopIteration:
                        raise exception
                    time.sleep(delay)
                else:
                    return result

        return _impl

    return _retry


class ExponentialBackoff(Backoff):
    """
    A backoff strategy with an exponentially increasing delay in between attempts.
    """

    def exponential(self, attempt: int) -> float:
        return min(self.maximum, pow(2, attempt) * self.start)

    def backoff(self, attempt: int) -> float:
        return self.exponential(attempt)


class ExponentialBackoffJitter(ExponentialBackoff):
    """
    An exponential backoff strategy with an added jitter that randomly spreads out the delays
    uniformly.
    """

    def backoff(self, attempt: int) -> float:
        return random.uniform(0, self.exponential(attempt))



import json
import time
from typing import Callable, List, Optional, Sequence, Union, cast
from uuid import UUID

from box import Box

from pycarlo.common import get_logger
from pycarlo.common.utils import boxify
from pycarlo.core import Client, Mutation, Query
from pycarlo.features.circuit_breakers.exceptions import (
    CircuitBreakerPipelineException,
    CircuitBreakerPollException,
)
from pycarlo.lib.schema import CircuitBreakerState, SqlJobCheckpointStatus

logger = get_logger(__name__)


class CircuitBreakerService:
    _TERM_STATES = {"PROCESSING_COMPLETE", "HAS_ERROR"}

    def __init__(
        self,
        mc_client: Optional[Client] = None,
        print_func: Callable = logger.info,
    ):
        """
        Convenience methods to help with using circuit breaker rules.

        :param mc_client: MCD client (e.g. for creating a custom session); created otherwise.
        :param print_func: Function to use for echoing. Uses python logging by default, which
                           requires setting MCD_VERBOSE_ERRORS.
        """
        self._client = mc_client or Client()
        self._print_func = print_func

    def trigger_and_poll(
        self,
        rule_uuid: Optional[Union[str, UUID]] = None,
        namespace: Optional[str] = None,
        rule_name: Optional[str] = None,
        timeout_in_minutes: int = 5,
    ) -> Optional[bool]:
        """
        Convenience method to both trigger and poll (wait) on circuit breaker rule execution.

        :param rule_uuid: UUID of the rule (custom SQL monitor) to execute.
        :param namespace: namespace of the rule (custom SQL monitor) to execute.
        :param rule_name: name of the rule (custom SQL monitor) to execute.
        :param timeout_in_minutes: Polling timeout in minutes. See poll() for details.
        :return: True if rule execution has breach; False otherwise. See poll() for any
                 exceptions raised.
        """
        breaches = self.poll_all(
            job_execution_uuids=self.trigger_all(
                rule_uuid=rule_uuid,
                namespace=namespace,
                rule_name=rule_name,
            ),
            timeout_in_minutes=timeout_in_minutes,
        )
        return bool(breaches > 0)

    def trigger(
        self,
        rule_uuid: Optional[Union[str, UUID]] = None,
        namespace: Optional[str] = None,
        rule_name: Optional[str] = None,
    ) -> str:
        """
        Trigger a rule to start execution with circuit breaker checkpointing.

        :param rule_uuid: UUID of the rule (custom SQL monitor) to execute.
        :param namespace: namespace of the rule (custom SQL monitor) to execute.
        :param rule_name: name of the rule (custom SQL monitor) to execute.
        :return: Job execution UUID, as a string, to be used to retrieve execution state / status.
        """
        mutation = Mutation()

        if rule_uuid:
            mutation.trigger_circuit_breaker_rule(rule_uuid=str(rule_uuid)).__fields__(
                "job_execution_uuid"
            )
        elif rule_name:
            if namespace:
                mutation.trigger_circuit_breaker_rule(
                    namespace=namespace, rule_name=rule_name
                ).__fields__("job_execution_uuid")
            else:
                mutation.trigger_circuit_breaker_rule(rule_name=rule_name).__fields__(
                    "job_execution_uuid"
                )
        else:
            raise ValueError("rule UUID or namespace and rule name must be specified")

        mutation_client = self._client(mutation)
        job_execution_uuid = mutation_client.trigger_circuit_breaker_rule.job_execution_uuid
        self._print_func(
            f"Triggered rule with ID '{rule_uuid}'. "
            f"Received '{job_execution_uuid}' as execution ID."
        )
        return cast(str, job_execution_uuid)

    def trigger_all(
        self,
        rule_uuid: Optional[Union[str, UUID]] = None,
        namespace: Optional[str] = None,
        rule_name: Optional[str] = None,
    ) -> List[str]:
        """
        Trigger a rule to start execution with circuit breaker checkpointing.

        This function supports rules that create multiple executions (e.g. rules with variables
        or over multiple tables)

        :param rule_uuid: UUID of the rule (custom SQL monitor) to execute.
        :param namespace: namespace of the rule (custom SQL monitor) to execute.
        :param rule_name: name of the rule (custom SQL monitor) to execute.
        :return: Job execution UUIDs, as strings, to be used to retrieve execution state / status.
        """
        mutation = Mutation()

        if rule_uuid:
            mutation.trigger_circuit_breaker_rule_v2(rule_uuid=str(rule_uuid)).__fields__(
                "job_execution_uuids"
            )
        elif rule_name:
            if namespace:
                mutation.trigger_circuit_breaker_rule_v2(
                    namespace=namespace, rule_name=rule_name
                ).__fields__("job_execution_uuids")
            else:
                mutation.trigger_circuit_breaker_rule_v2(rule_name=rule_name).__fields__(
                    "job_execution_uuids"
                )
        else:
            raise ValueError("rule UUID or namespace and rule name must be specified")

        job_execution_uuids = [
            str(id)
            for id in self._client(mutation).trigger_circuit_breaker_rule_v2.job_execution_uuids
        ]
        self._print_func(
            f"Triggered rule with ID '{rule_uuid}'. "
            f"Received {job_execution_uuids} as execution IDs."
        )
        return job_execution_uuids

    def poll(
        self,
        job_execution_uuid: Union[str, UUID],
        timeout_in_minutes: int = 5,
    ) -> Optional[int]:
        """
        Poll status / state of an execution for a triggered rule. Polls until status is in a term
        state or timeout.

        :param job_execution_uuid: UUID for the job execution of a rule (custom SQL monitor).
        :param timeout_in_minutes: Polling timeout in minutes. Note that The Data Collector Lambda
                                   has a max timeout of 15 minutes when executing a query. Queries
                                   that take longer to execute are not supported, so we recommend
                                   filtering down the query output to improve performance (e.g limit
                                   WHERE clause). If you expect a query to take the full 15 minutes
                                   we recommend padding the timeout to 20 minutes.
        :return: Breach count across all executions. A greater than 0 value indicates a breach.
        :raise CircuitBreakerPipelineException: An error in executing the
                                                rule (e.g. error in query).
        :raise CircuitBreakerPollException: A timeout during polling or a malformed response.
        """
        return self.poll_all([job_execution_uuid], timeout_in_minutes=timeout_in_minutes)

    def poll_all(
        self,
        job_execution_uuids: Sequence[Union[str, UUID]],
        timeout_in_minutes: int = 5,
    ) -> int:
        """
        Poll status / state of executions for a triggered rule. Polls until status is in a term
        state or timeout.

        :param job_execution_uuids: UUIDs for the job executions of a rule (custom SQL monitor).
        :param timeout_in_minutes: Polling timeout in minutes. Note that The Data Collector Lambda
                                   has a max timeout of 15 minutes when executing a query. Queries
                                   that take longer to execute are not supported, so we recommend
                                   filtering down the query output to improve performance (e.g limit
                                   WHERE clause). If you expect a query to take the full 15 minutes
                                   we recommend padding the timeout to 20 minutes.
        :return: Breach count. A greater than 0 value indicates a breach.
        :raise CircuitBreakerPipelineException: An error in executing the
                                                rule (e.g. error in query).
        :raise CircuitBreakerPollException: A timeout during polling or a malformed response.
        """
        logs = cast(
            List[Box],
            self._poll(
                job_execution_uuids=job_execution_uuids,
                timeout_in_minutes=timeout_in_minutes,
            ),
        )
        self._print_func(
            "Completed polling. Retrieved execution with logs "
            f"{list(map(str, logs))} for IDs {job_execution_uuids}."
        )

        breaches = 0
        has_breaches = False
        if logs and len(logs) > 0:
            for log in logs:
                if log.payload.error:
                    logs_str = "\n".join(str(log) for log in logs)
                    raise CircuitBreakerPipelineException(
                        f"Execution pipeline errored out. Details:\n{logs_str}"
                    )
                if log.payload.breach_count is not None:
                    breaches += log.payload.breach_count
                    has_breaches = True

        if not has_breaches:
            raise CircuitBreakerPollException

        return breaches

    @boxify(use_snakes=True, default_box_attr=None, default_box=True)
    def _poll(
        self,
        job_execution_uuids: Sequence[Union[str, UUID]],
        timeout_in_minutes: int,
        sleep_interval_in_seconds: int = 15,
    ) -> Optional[List[Box]]:
        timeout_start = time.time()
        while time.time() < timeout_start + 60 * timeout_in_minutes:
            query = Query()
            query.get_circuit_breaker_rule_state_v2(
                job_execution_uuids=map(str, job_execution_uuids)
            ).__fields__("status", "log")
            circuit_rule_breaker_states = cast(
                List[CircuitBreakerState],
                self._client(query).get_circuit_breaker_rule_state_v2,
            )

            aggregated_status = self._get_aggregated_status(circuit_rule_breaker_states)
            self._print_func(
                f"Retrieved execution with aggregated status '{aggregated_status}' for "
                f"IDs {job_execution_uuids}."
            )

            if aggregated_status in self._TERM_STATES:
                return self._get_payloads(circuit_rule_breaker_states, aggregated_status)

            self._print_func(
                f"Aggregated state is not terminal state for IDs {job_execution_uuids}. "
                f"Polling again in '{sleep_interval_in_seconds}' seconds."
            )
            time.sleep(sleep_interval_in_seconds)

    def _get_log_payload(self, log: str):
        log_entries = json.loads(log)
        log_entries.reverse()
        for entry in log_entries:
            if "payload" in entry:
                return Box(entry, default_box_attr=None, default_box=True)
        return Box()

    def _get_payloads(
        self,
        states: List[CircuitBreakerState],
        status: SqlJobCheckpointStatus,
    ) -> List[Box]:
        payloads = []
        for state in states:
            if state.status == status:
                payloads.append(self._get_log_payload(str(state.log)))
        return payloads

    @staticmethod
    def _get_aggregated_status(states: List[CircuitBreakerState]) -> SqlJobCheckpointStatus:
        if not states:
            return SqlJobCheckpointStatus.REGISTERED  # type: ignore

        status_by_state = {}
        for state in states:
            status_by_state.setdefault(state.status, []).append(state)

        def all_in_state(s: SqlJobCheckpointStatus):
            return len(status_by_state.get(s, [])) == len(states)

        return (
            SqlJobCheckpointStatus.HAS_ERROR  # type: ignore
            if status_by_state.get(SqlJobCheckpointStatus.HAS_ERROR)  # type: ignore
            else SqlJobCheckpointStatus.PROCESSING_COMPLETE  # type: ignore
            if all_in_state(SqlJobCheckpointStatus.PROCESSING_COMPLETE)  # type: ignore
            else SqlJobCheckpointStatus.PROCESSING_START  # type: ignore
            if status_by_state.get(SqlJobCheckpointStatus.PROCESSING_COMPLETE)  # type: ignore
            or all_in_state(SqlJobCheckpointStatus.PROCESSING_START)  # type: ignore
            else SqlJobCheckpointStatus.EXECUTING_COMPLETE  # type: ignore
            if all_in_state(SqlJobCheckpointStatus.PROCESSING_COMPLETE)  # type: ignore
            else SqlJobCheckpointStatus.EXECUTING_START  # type: ignore
            if status_by_state.get(SqlJobCheckpointStatus.EXECUTING_COMPLETE)  # type: ignore
            or all_in_state(SqlJobCheckpointStatus.EXECUTING_START)  # type: ignore
            else SqlJobCheckpointStatus.REGISTERED  # type: ignore
        )



"""
Generates JSON schema from introspection. Use `make generate`.
"""
import json

from sgqlc.endpoint.requests import RequestsEndpoint
from sgqlc.introspection import query, variables
from sgqlc.introspection.__main__ import get_arg_parse

from utils.vars import ensure_token_env_vars

if __name__ == "__main__":
    api_id, api_token = ensure_token_env_vars()
    args = get_arg_parse().parse_args()

    headers = {"x-mcd-id": api_id, "x-mcd-token": api_token}
    endpoint = RequestsEndpoint(url=args.url, base_headers=headers)
    data = endpoint(query, variables(include_description=True, include_deprecated=True))

    if data.get("errors"):
        raise SystemExit(f"Failed to retrieve schema with - {data.get('errors')}")
    json.dump(data, args.outfile, sort_keys=True, indent=2, default=str)
